#include <Arduino.h>
#include <OneWire.h>
#include <DallasTemperature.h>
#include <Wire.h>

#include "buzzer.h"
#include "DS1340.h"
#include "ht1632.h"
#include "font5x7.h"
#include "displaytiles.h"

#define pinCountdownKey  8
#define pinSetupKey		12
#define pinPlusKey		11
#define pinMinusKey		 9
#define pinBuzzer		5

typedef struct
{
	int pin;
	bool keystate;
	int keyCode;
        byte keyOptions;
        #define KEY_OPT_NORMAL    0x00
        #define KEY_OPT_REPEAT    0x01
        unsigned long keyTime;
} KEYS_Vector;

KEYS_Vector keysVector[] = 
	{ 
		{ pinCountdownKey,	1, KEYCODE_TIMER,	KEY_OPT_NORMAL, 0L },
		{ pinSetupKey,		1, KEYCODE_SET,		KEY_OPT_NORMAL, 0L },
		{ pinPlusKey,		1, KEYCODE_PLUS,	KEY_OPT_REPEAT, 0L },	// Long press interepreted as repeatitive press
		{ pinMinusKey,		1, KEYCODE_MINUS,	KEY_OPT_REPEAT, 0L },	// Long press interepreted as repeatitive press
		{ 0, 0, 0, 0, 0L }
	};
	


void SetupButtons()
{
	int i;
	
	for (i = 0; keysVector[i].pin != 0; i++)
	{
		pinMode(keysVector[i].pin, INPUT);
		digitalWrite(keysVector[i].pin, HIGH); // This should enable the pull up resistors
		keysVector[i].keyTime = 0L;
	}
}


void setup()
{
	Serial.begin(9600);
	Serial.println("Starting...");

	Serial.println("Setting buzzer");
	buzzer.init(pinBuzzer);
	buzzer.playStartTune();

	Serial.println("Initializing:\n\tLED Matrix");
	ledMatrixDisplay.init();

	Serial.println("\tI2C");
	Wire.begin();

	Serial.println("\tRTC");
	rtcModule.init();

	Serial.println("\tDisplay manager");
	displayManager.init();

	Serial.println("Setting buttons");
	SetupButtons();

	Serial.println("setup complete");
}


int ReadKeys()
{
	#define KEY_OPT_REPEAT_INIT	500L	// Only if key is held longer than this, we will start emulating key pressing
	#define KEY_OPT_REPEAT_TIME	200L	// Time between emulating key pressing 

	int i, key, rv = 0;

	for (i = 0; keysVector[i].pin != 0; i++)
	{
		key = digitalRead(keysVector[i].pin );
		if (keysVector[i].keyOptions == KEY_OPT_NORMAL)
			if ( (key == 0) && (keysVector[i].keystate == 1) ) // First time we see button pushed
				rv = keysVector[i].keyCode;
			else;
		else
		{
			// Needs to emulate repetitive key presses
			unsigned long _now = millis();

			if (key == 0) // Key is now pressed)
				if (keysVector[i].keyTime == 0L)
				{
					// First time it was pressed, store (press time + initial delay) and return keyCode
					keysVector[i].keyTime = _now + KEY_OPT_REPEAT_INIT - KEY_OPT_REPEAT_TIME;
					rv = keysVector[i].keyCode;
				}
				else
				{
					if (((_now  > keysVector[i].keyTime) && (_now - keysVector[i].keyTime) > KEY_OPT_REPEAT_TIME))
					{
						keysVector[i].keyTime = _now; // Presereve this time
						rv = keysVector[i].keyCode;
					}
					else	// Not enough time to generate a repeat
                                            key = 1;
				}
			else
			{
				// Key is not pressed, reset the time
				keysVector[i].keyTime = 0L;
			}
		}
		keysVector[i].keystate = key;
		if (rv)
			break;
	}
	return rv;
}


void loop ()
{
	static unsigned long last = 0;
	unsigned long _now = millis();
	
	if ((_now - last) > 1000UL)
	{
		displayManager.tick();
		last = millis();
	}
	displayManager.show();
	int keyCode = ReadKeys();
	if (keyCode != 0)
		displayManager.sendKey(keyCode);

}
