/* ===========================================================================================
 *	Class to control Sure Electronics Bi-Color (RED/GREEN) 32x16 LED Matrix
 * =========================================================================================== */
#include <avr/pgmspace.h>
#include <Arduino.h>

#include "ht1632.h"
#include "font5x7.h"

// -------------------------------------------------------------------------------------------
// Function Mame: HT1632ColorDisplay
// Purpose: Construct object of display with setting pin numbers to defaults
// -------------------------------------------------------------------------------------------
HT1632ColorDisplay::HT1632ColorDisplay()
{
	ht1632_data = 7;
	ht1632_clk = 2;
	ht1632_wrclk = 3;
	ht1632_cs = 4;

	fontWidth = 5;
}

// -------------------------------------------------------------------------------------------
// Function Mame: HT1632ColorDisplay
// Purpose: Construct object of display with custom pin numbers
// Parameters:	data	The DATA pin
//				clk		The CLK pin (clock)
//				wrclk	The WRITE CLOCK pin
//				cs		The CS pin (chip-select)
// -------------------------------------------------------------------------------------------
HT1632ColorDisplay::HT1632ColorDisplay(byte data, byte clk, byte wrclk, byte cs)
{
	ht1632_data = data;
	ht1632_clk = clk;
	ht1632_wrclk = wrclk;
	ht1632_cs = cs;
	
	fontWidth = 5;
}




// -------------------------------------------------------------------------------------------
// Function Name: OutputCLK_Pulse
// Purpose: enable CLK_74164 pin to output a clock pulse
// -------------------------------------------------------------------------------------------
void HT1632ColorDisplay::OutputCLK_Pulse(void)	//Output a clock pulse
{
	digitalWrite(ht1632_clk, HIGH);
	digitalWrite(ht1632_clk, LOW);
}

// -------------------------------------------------------------------------------------------
// Function Name: chipselect
// Purpose: enable HT1632C
// Parameters:
//			select	HT1632C to be selected
//					If select=0, select none.
//					If s<0, select all.
// -------------------------------------------------------------------------------------------
void HT1632ColorDisplay::chipselect(int select)
{
	unsigned char tmp = 0;
	if (select < 0)
	{ // Enable all HT1632C
		digitalWrite(ht1632_cs, LOW);
		for (tmp=0; tmp<CHIP_MAX; tmp++)
		{
			OutputCLK_Pulse();
		}
	}
	else
	if(select==0)
	{
		//Disable all HT1632Cs
		digitalWrite(ht1632_cs, HIGH);
		for(tmp=0; tmp<CHIP_MAX; tmp++)
		{
			OutputCLK_Pulse();
		}
	}
	else
	{
		digitalWrite(ht1632_cs, HIGH);
		for(tmp=0; tmp<CHIP_MAX; tmp++)
		{
			OutputCLK_Pulse();
		}
		digitalWrite(ht1632_cs, LOW);
		OutputCLK_Pulse();
		digitalWrite(ht1632_cs, HIGH);
		tmp = 1;
		for( ; tmp<select; tmp++)
		{
			OutputCLK_Pulse();
		}
	}
}
// -------------------------------------------------------------------------------------------
// Function Name: writebits
// Purpose: Write bits (up to 8) to h1632 on pins ht1632_data, ht1632_wrclk
//                  Chip is assumed to already be chip-selected
//                  Bits are shifted out from MSB to LSB, with the first bit sent
//                  being (bits & firstbit), shifted till firsbit is zero.
// Parameters:
//		bits		Bits to send
//	    firstbit	The first bit to send
// -------------------------------------------------------------------------------------------
void writebits (byte bits, byte firstbit)
{
  while (firstbit)
  {
    digitalWrite(3, LOW);
    if (bits & firstbit)
	digitalWrite(7, HIGH);
    else
	digitalWrite(7, LOW);
    digitalWrite(3, HIGH);
    firstbit >>= 1;
  }
}

void HT1632ColorDisplay::_writebits (byte bits, byte firstbit)
{
	while (firstbit)
	{
		digitalWrite(ht1632_wrclk, LOW);
		if (bits & firstbit)
			digitalWrite(ht1632_data, HIGH);
		else
			digitalWrite(ht1632_data, LOW);
		digitalWrite(ht1632_wrclk, HIGH);
		firstbit >>= 1;
	}
}

// -------------------------------------------------------------------------------------------
// Function Name: sendcmd
// Purpose: Send a command to the ht1632 chip.
//                  Select 1 0 0 c7 c6 c5 c4 c3 c2 c1 c0 xx Free
// Parameters:
//		chipNo	The chip you want to send data
//      command	Consists of a 3-bit "CMD" ID, an 8bit command, and one "don't care bit".
// -------------------------------------------------------------------------------------------
void HT1632ColorDisplay::sendcmd (byte chipNo, byte command)
{
	chipselect(chipNo);
	writebits(HT1632_ID_CMD, 1<<2);  // send 3 bits of id: COMMMAND
	writebits(command, 1<<7);  // send the actual command
	writebits(0, 1); 	/* one extra dont-care bit in commands. */
	chipselect(0);
}

// -------------------------------------------------------------------------------------------
// Function Name: senddata
// Purpose: send a nibble (4 bits) of data to a particular memory location of the
//                  ht1632.  The command has 3 bit ID, 7 bits of address, and 4 bits of data.
//                  Select 1 0 1 A6 A5 A4 A3 A2 A1 A0 D0 D1 D2 D3 Free
//                  Note that the address is sent MSB first, while the data is sent LSB first!
//                  This means that somewhere a bit reversal will have to be done to get
//                  zero-based addressing of words and dots within words.
// Parameters:
//		chipNo		The chip you want to send data
//		address		Chip address to write to
//		data		Data to write to chip memory
// -------------------------------------------------------------------------------------------
void HT1632ColorDisplay::senddata (byte chipNo, byte address, byte data)
{
	chipselect(chipNo);
	writebits(HT1632_ID_WR, 1<<2);  // send ID: WRITE to RAM
	writebits(address, 1<<6); // Send address
	writebits(data, 1<<3); // send 4 bits of data
	chipselect(0);
}

// -------------------------------------------------------------------------------------------
// Function Name: _clear
// Purpose: clear display including the shadow memory for region of chips c0 to c1
// Parameters:
//		c0	The first chip (1 is the first one)
//		c1	The last chip to clear
// -------------------------------------------------------------------------------------------
void HT1632ColorDisplay::_clear(int c0, int c1)
{
  byte i;
  
	for (int j = c0; j <= c1; j++)
	{
		chipselect(j);
		writebits(HT1632_ID_WR, 1<<2);  // send ID: WRITE to RAM
		writebits(0, 1<<6);				// Send address
		for (i=0; i<64/2; i++)			// Clear entire display
			writebits(0, 1<<7);			// send 8 bits of data
		chipselect(0);
		for (i=0; i<64; i++)
			ht1632_shadowram[i][j-1] = 0;
	}
}

// -------------------------------------------------------------------------------------------
// Function Name: clear
// Purpose: clear entire display including the shadow memory
// -------------------------------------------------------------------------------------------
void HT1632ColorDisplay::clear()
{
	_clear(1, CHIP_MAX);
}


// -------------------------------------------------------------------------------------------
// Function Name: clearLine
// Purpose: clear one line in the display including the shadow memory
// -------------------------------------------------------------------------------------------
void HT1632ColorDisplay::clearLine(int line)
{
	int startChip, endChip;
	
	if (line == 1)
	{
		startChip = 1;
		endChip = 2;
	}
	else
	{
		startChip = 3;
		endChip = 4;
	}	
	_clear(startChip, endChip);
}

// -------------------------------------------------------------------------------------------
// Function Name: setIntensity
// Purpose: Sets the intensity of any of the quadrant of the display
// Parameters:
//		chip		The chip responsible for the quadrant [1..4]
//		intensity	The intensity value [0..15]
// -------------------------------------------------------------------------------------------
void HT1632ColorDisplay::setIntensity(int chip, int intensity)
{
	byte i = intensity & 0x0F;
	
	i|= 0xA0;
	sendcmd(chip, i); 
}


// -------------------------------------------------------------------------------------------
// Function Name: setIntensity
// Purpose: Sets the intensity of any of the quadrant of the display
// Parameters:
//		intensity	The intensity value [0..15]
// -------------------------------------------------------------------------------------------
void HT1632ColorDisplay::setIntensity(int intensity)
{
	byte i;
	
	for (i = 1; i <= CHIP_MAX; i++)
		setIntensity(i, intensity);
}


// -------------------------------------------------------------------------------------------
// Function Name: xyToIndex
// Purpose: get the value of x,y
// Parameters:
//		x		X coordinate
//		y		Y coordinate
//Return Code: address of xy pixel
// -------------------------------------------------------------------------------------------
byte HT1632ColorDisplay::xyToIndex(byte x, byte y)
{
	byte nChip, addr;
	
	if (x>=32)
		nChip = 3 + x/16 + (y>7?2:0);
	else
		nChip = 1 + x/16 + (y>7?2:0);
	
	x = x % 16;
	y = y % 8;
	addr = (x<<1) + (y>>2);
	
	return addr;
}

// -------------------------------------------------------------------------------------------
// Function Name: calcBit
// Purpose: calculate the bitval of y
// Parameters:
//		y		Y coordinate
// Return Code: bitval
// -------------------------------------------------------------------------------------------
#define calcBit(y) (8>>(y&3))

// -------------------------------------------------------------------------------------------
// Function Name: getPixel
// Purpose: get the value (color) of pixel in x,y. Function is using the shadow memory and does
//		not read the actual value from the display, though possible
// Parameters:
//		x		X coordinate
//      y		Y coordinate
// Return Code: color setted on x,y coordinates
// -------------------------------------------------------------------------------------------
int HT1632ColorDisplay::getPixel(byte x, byte y)
{
	byte addr, bitval, nChip;
	
	if (x>=32)
		nChip = 3 + x/16 + (y>7?2:0);
	else
		nChip = 1 + x/16 + (y>7?2:0);
	
	addr = xyToIndex(x,y);
	bitval = calcBit(y);
	
	if ((ht1632_shadowram[addr][nChip-1] & bitval) && (ht1632_shadowram[addr+32][nChip-1] & bitval))
		return ORANGE;
	else
		if (ht1632_shadowram[addr][nChip-1] & bitval)
			return GREEN;
		else
			if (ht1632_shadowram[addr+32][nChip-1] & bitval)
				return RED;
			else
				return BLACK;
}


// -------------------------------------------------------------------------------------------
// Function Name: plot
// Purpose: plot a dot/pixel on x,y with given color
// Parameters:
//		x		X coordinate
//      y		Y coordinate
//      color	The color to show, BLACK(clean), GREEN, RED, ORANGE
// -------------------------------------------------------------------------------------------
void HT1632ColorDisplay::plot (byte x, byte y, byte color)
{
	byte nChip, addr, bitval;
	
	if (x<0 || x>X_MAX || y<0 || y>Y_MAX)
		return;
	
	if (color != BLACK && color != GREEN && color != RED && color != ORANGE)
		return;
	
	if (x>=32)
		nChip = 3 + x/16 + (y>7?2:0);
	else
		nChip = 1 + x/16 + (y>7?2:0);
	
	addr = xyToIndex(x,y);
	bitval = calcBit(y);

	switch (color)
	{
		case BLACK:
		if (getPixel(x,y) != BLACK)
		{
			// compare with memory to only set if pixel is other color
			// clear the bit in both planes;
			ht1632_shadowram[addr][nChip-1] &= ~bitval;
			senddata(nChip, addr, ht1632_shadowram[addr][nChip-1]);
			addr = addr + 32;
			ht1632_shadowram[addr][nChip-1] &= ~bitval;
			senddata(nChip, addr, ht1632_shadowram[addr][nChip-1]);
		}
		break;
		case GREEN:
		if (getPixel(x,y) != GREEN)
		{
			// compare with memory to only set if pixel is other color
			// set the bit in the green plane and clear the bit in the red plane;
			ht1632_shadowram[addr][nChip-1] |= bitval;
			senddata(nChip, addr, ht1632_shadowram[addr][nChip-1]);
			addr = addr + 32;
			ht1632_shadowram[addr][nChip-1] &= ~bitval;
			senddata(nChip, addr, ht1632_shadowram[addr][nChip-1]);
		}
		break;
		case RED:
		if (getPixel(x,y) != RED)
		{
			// compare with memory to only set if pixel is other color
			// clear the bit in green plane and set the bit in the red plane;
			ht1632_shadowram[addr][nChip-1] &= ~bitval;
			senddata(nChip, addr, ht1632_shadowram[addr][nChip-1]);
			addr = addr + 32;
			ht1632_shadowram[addr][nChip-1] |= bitval;
			senddata(nChip, addr, ht1632_shadowram[addr][nChip-1]);
		}
		break;
		case ORANGE:
		if (getPixel(x,y) != ORANGE)
		{
			// compare with memory to only set if pixel is other color
			// set the bit in both the green and red planes;
			ht1632_shadowram[addr][nChip-1] |= bitval;
			senddata(nChip, addr, ht1632_shadowram[addr][nChip-1]);
			addr = addr + 32;
			ht1632_shadowram[addr][nChip-1] |= bitval;
			senddata(nChip, addr, ht1632_shadowram[addr][nChip-1]);
		}
		break;
	}
}

// -------------------------------------------------------------------------------------------
// Function Name: init
// Purpose: initialize the display by assigning the pin ports and setting up parameters to the
//	display itself. On return the display should be clear
// -------------------------------------------------------------------------------------------
void HT1632ColorDisplay::init()
{
	pinMode(ht1632_cs, OUTPUT);
	digitalWrite(ht1632_cs, HIGH);
	pinMode(ht1632_wrclk, OUTPUT);
	pinMode(ht1632_data, OUTPUT);
	pinMode(ht1632_clk, OUTPUT);
	for (int i=1; i<=CHIP_MAX; i++)
	{
		sendcmd(i, HT1632_CMD_SYSDIS); // Disable system
		sendcmd(i, HT1632_CMD_COMS00); // 16*32, PMOS drivers
		sendcmd(i, HT1632_CMD_MSTMD);  // MASTER MODE
		sendcmd(i, HT1632_CMD_SYSON);  // System on
		sendcmd(i, HT1632_CMD_LEDON);  // LEDs on
		setIntensity(i, 0x0F);
	}
	clear(); // Clear the display
}



// -------------------------------------------------------------------------------------------
// Function Name: print
// Purpose:	prints a string on the display at given line and color
// Parameters:
//		s		The string to display
//		line	On which line to print [1..2]
//		color	The color to use [BLACK, GREEN, RED, ORGANGE]
// -------------------------------------------------------------------------------------------
void HT1632ColorDisplay::print(const char *s, int line, int color)
{
    char c;
    int x, i, j;
    int offset = (line == 1) ? 0 : 8;
    int mask;
    
    for(x=0; s[x]; x++)
    {
        c = s[x];
        clearFontCodeBuffer();
        fillFontCodeBuffer(c);
        for(i=0; i <8;i++)
        {
            mask = 1 << i;
            for(j=0; j < fontWidth;j++)
            {
                if (fontCode[j] & mask)
                    plot(j+((fontWidth+1)*x), (7-i)+offset, color);
                else
                    plot(j+((fontWidth+1)*x), (7-i)+offset, BLACK);
            }
        }
    }
}


void HT1632ColorDisplay::print(int startX, int line, const char *s, int color, int cwidth)
{
    char c;
    int x, i, j;
    int offset = (line == 1) ? 0 : 8;
    int mask;
    
    for(x=0; s[x]; x++)
    {
        c = s[x];
        clearFontCodeBuffer();
        fillFontCodeBuffer(c, cwidth);
        for(i=0; i <8;i++)
        {
            mask = 1 << i;
            for(j=0; j < cwidth;j++)
            {
                if (fontCode[j] & mask)
                    plot(j+((cwidth+1)*x)+startX, (7-i)+offset, color);
                else
                    plot(j+((cwidth+1)*x)+startX, (7-i)+offset, BLACK);
            }
        }
    }
}





// -------------------------------------------------------------------------------------------
// Function Name: fillFontCodeBuffer
// Purpose: loads a charcter font code int local buffer
// Parameters:
//		chr		The desired character
// -------------------------------------------------------------------------------------------
void HT1632ColorDisplay::fillFontCodeBuffer(char chr, int cwidth)
{
  if (cwidth == 3)
    for(int i=0; _myTinyCharacters[i]; i++)
    {
        if(_myTinyCharacters[i] == chr)
        {
            memcpy(fontCode, _myTinyFont[i], cwidth);
            return;
        }
    }
   else
    for(int i=0; _myfontCharacters[i]; i++)
    {
        if(_myfontCharacters[i] == chr)
        {
          memcpy(fontCode, _myfont[i], cwidth);
            return;
        }
    }
}

// -------------------------------------------------------------------------------------------
// Function Name: clearFontCodeBuffer
// Purpose: Clears the font code in local buffer
// -------------------------------------------------------------------------------------------
void HT1632ColorDisplay::clearFontCodeBuffer()
{
	memset(fontCode, 0, sizeof(fontCode));
}


HT1632ColorDisplay	ledMatrixDisplay;